#ifndef UNCONSTRAINEDVARIABLESIMPLIFIER_H
#define UNCONSTRAINEDVARIABLESIMPLIFIER_H

#include "z3++.h"
#include <map>
#include <vector>
#include <string>
#include <tuple>

enum BoundType { EXISTENTIAL, UNIVERSAL };

class UnconstrainedVariableSimplifier
{
public:
    UnconstrainedVariableSimplifier(z3::context &ctx, z3::expr expr) : expression(expr)
    {
      this->context = &ctx;

      std::vector<BoundVar> boundVars;
      variableCounts = countVariableOccurences(expression, boundVars).first;
    }

    void PrintUnconstrained()
    {
        bool allConstrained = true;

        for (auto &item : variableCounts)
        {
            std::cout << "var " << item.first << " - " << item.second << " times" << std::endl;

            if (item.second == 1)
            {
                allConstrained = false;
                //std::cout << "Unconstrained variable: " << item.first << std::endl;
            }
        }
        if (allConstrained) std::cout << "All variables constrained" << std::endl;
    }

    void SimplifyOnce()
    {
        expression = simplifyOnce(expression, {}, true);
    }

    z3::expr GetExpr() const { return expression; }

    void SimplifyIte();
    z3::expr CanonizeBoundVariables(const z3::expr&);

private:
    z3::context* context;
    z3::expr expression;

    typedef std::tuple<std::string, BoundType, int> BoundVar;

    std::map<const Z3_ast, std::pair<std::map<std::string, int>, std::vector<BoundVar>>> subformulaVariableCounts;
    std::map<const Z3_ast, int> subformulaMaxDeBruijnIndices;
    std::map<std::string, int> variableCounts;

    typedef std::map<const Z3_ast, std::pair<z3::expr, const std::vector<BoundVar>>> cacheMapType;

    cacheMapType trueSimplificationCache;
    cacheMapType falseSimplificationCache;

    std::pair<std::map<std::string, int>, int> countVariableOccurences(z3::expr, std::vector<BoundVar>);

    z3::expr simplifyOnce(z3::expr, std::vector<BoundVar>, bool);
    bool isUnconstrained(z3::expr, const std::vector<BoundVar>&);
    bool isVar(z3::expr);
    bool isBefore(z3::expr, z3::expr);
    BoundType getBoundType(z3::expr, const std::vector<BoundVar>&);

    int getNumberOfLeadingZeroes(const z3::expr&);
    int lastBound = 0;
};

#endif // UNCONSTRAINEDVARIABLESIMPLIFIER_H
