
#include <iostream>
#include <string>
#include <z3++.h>
#include <cmath>
#include <fstream>
#include <getopt.h>

#include <chrono>

#include "UnconstrainedVariableSimplifier.h"

using namespace std;
using namespace z3;

using std::chrono::high_resolution_clock;
using std::chrono::milliseconds;

int main(int argc, char* argv[])
{
    static struct option long_options[] = {
	{"propagate-unconstrained", no_argument, 0, 'p' },
	{"canonize", no_argument, 0, 'c' },
	{"run-z3", no_argument, 0, 'z' },
	{0,           0,                 0,  0   }
    };

    bool propagateUnconstrainedFlag = false, runZ3 = false, canonize = false;
    char* filename;

    int opt = 0;

    int long_index = 0;
    while ((opt = getopt_long(argc, argv,"pzc", long_options, &long_index )) != -1)
    {
	switch (opt)
	{
	case 'p':
	    propagateUnconstrainedFlag = true;
	    break;
	case 'z':
	    runZ3 = true;
	    break;
	case 'c':
	    canonize = true;
	    break;
	default:
	    std::cout << "Invalid arguments" << std::endl;
	    exit(1);
	}
    }

    if (optind < argc)
    {
	filename = argv[optind];
    }
    else
    {
	std::cout << "Filename required" << std::endl;
	return 1;
    }

    z3::context ctx;
    Z3_ast ast = Z3_parse_smtlib2_file(ctx, filename, 0, 0, 0, 0, 0, 0);
    expr e = to_expr(ctx, ast);

    e = e.simplify();

    if (canonize)
    {
	UnconstrainedVariableSimplifier simplifier(ctx, e);
	e = simplifier.CanonizeBoundVariables(e);
    }

    if (propagateUnconstrainedFlag)
    {
	UnconstrainedVariableSimplifier simplifier(ctx, e);
	//simplifier.PrintUnconstrained();
	simplifier.SimplifyIte();
	e = simplifier.GetExpr();
    }

    e = e.simplify();
    //std::cout << e << std::endl;

    if (runZ3)
    {
	solver s(ctx);
	s.add(e);
	params p(ctx);
	p.set("smt.timeout", 5000u);
	s.set(p);
	std::cout << s.check() << std::endl;
    }
    else
    {
	std::cout << e << std::endl;
    }
}
